import re
from contextlib import suppress
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, Final, List, TypeAlias, cast  # noqa: UP035
from urllib.parse import unquote, urlparse

import lxml.html
from django.conf import settings
from django.db import models
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django_extensions.db.models import TimeStampedModel
from lxml import etree
from lxml.html import HtmlElement
from wildewidgets.models import ViewSetMixin

from .fields import MachineNameField
from .settings import MAX_GLOBAL_TOC_TREE_DEPTH
from .validators import NoHTMLValidator

F: TypeAlias = models.Field
M2M: TypeAlias = models.ManyToManyField
FK: TypeAlias = models.ForeignKey


# --------------------------
# Dataclasses
# --------------------------


@dataclass
class TreeNode:
    """
    A :py:class:`dataclass` that we use with :py:class:`SphinxPageTree`
    to build out the global navigation structure for a set of documentation for
    a :py:class:`Version`.
    """

    #: The page title
    title: str
    #: This page
    page: "SphinxPage | None" = None
    #: The :py:class:`SphinxPage` after this one
    prev: "SphinxPage | None" = None
    #: The :py:class:`SphinxPage` before this one
    next: "SphinxPage | None" = None
    #: The :py:class:`SphinxPage` that is this page's parent
    parent: "SphinxPage | None" = None
    #: The :py:class:`TreeNode` objects that are this page's children
    children: list["TreeNode"] = field(default_factory=list)

    @classmethod
    def from_page(cls, page: "SphinxPage") -> "TreeNode":
        """
        Build a :py:class:`TreeNode` from ``page``.

        Note:
            This does not populate :py:attr:`children`; :py:class:`SphinxPageTree`
            will populate it as appropriate as it ingests pages.

        Args:
            page: the :py:class:`SphinxPage` from which to build a node

        Returns:
            A configured node.

        """
        return cls(
            page=page,
            title=page.title,
            next=page.next_page,  # type: ignore[arg-type]
            prev=page.previous_page.first(),
            parent=page.parent,  # type: ignore[arg-type]
        )


@dataclass
class ClassifierNode:
    # The title of the classifier node
    title: str
    # The classifier for the classifier node
    classifier: "Classifier | None" = None
    # The items for the classifier node
    items: dict[str, "ClassifierNode"] = field(default_factory=dict)


# --------------------------
# Helper classes
# --------------------------


class SphinxPageTree:
    """
    A class that holds the page hierarchy for the set of :py:class:`SphinxPage`
    pages in a :py:class:`Version`.as a linked set of :py:class:`TreeNode`
    objects.

    The page heirarchy is built by starting at :py:attr:`Version.head` and
    following the page linkages by looking at :py:attr:`SphinxPage.next_page`,
    stopping the traversal when we find a :py:attr:`SphinxPage.next_page` that
    is ``None``.

    As we traverse, if a :py:attr:`SphinxPage.parent` is not ``None``, find the
    :py:class:`TreeNode` for that parent, and add the page to
    :py:attr:`TreeNode.children`.

    For pages who have no :py:attr:`SpinxPage.parent`, assume they are top level
    children of the set, and make them children of :py:class:`Version.head`.

    Load it like so::

        >>> project = Project.objects.get(machine_name='my-project')
        >>> version = project.versions.get(version='1.0.0')
        >>> tree = SphinxPageTree(version)

    You can then traverse the built hierarchy by starting at
    :py:attr:`SphinxPageTree.head`, looking at its children, then looking at
    their children, etc..

    Args:
        version: the :py:class:`Version` to build the tree for

    """

    def __init__(self, version: "Version") -> None:
        #: The :py:class:`Version` that this tree examines
        self.version: Version = version
        self.nodes: Dict[int, TreeNode] = {}
        self.nodes[self.version.head.id] = TreeNode.from_page(self.version.head)  # type: ignore[attr-defined, arg-type]
        #: The top page in the page hierarchy
        self.head: TreeNode = self.nodes[self.version.head.id]  # type: ignore[attr-defined]
        self.build(version)

    def build(self, version: "Version"):
        self.add_page(version.head)  # type: ignore[arg-type]

    def add_page(self, page: "SphinxPage"):
        node = TreeNode.from_page(page)
        self.nodes[page.id] = node
        if node.parent:
            if node.parent.id in self.nodes:
                self.nodes[node.parent.id].children.append(node)
        elif node != self.head:
            # The top level pages that are not head will not have any parent
            # because Sphinx doesn't think that way
            self.head.children.append(node)
        if node.next:
            self.add_page(node.next)

    def _traverse_level(self, pages, nodes):
        for node in nodes:
            if node.page:
                pages.append(node.page)
                if node.children:
                    self._traverse_level(pages, node.children)

    def traverse(self) -> list["SphinxPage"]:
        """
        Return a list of the pages represented in this tree.
        """
        pages: List["SphinxPage"] = [cast("SphinxPage", self.head.page)]  # noqa: UP037
        self._traverse_level(pages, self.head.children)
        return pages


class SphinxPageTreeProcessor:
    def build_item(self, node: TreeNode) -> Dict[str, Any]:
        """
        Build a :py:class:`wildewdigets.MenuItem` compatible
        dict representing ``node``.

        Args:
            node: the current node in our page tree

        Returns:
            A dict suitable for loading into a :py:class:`wildewidgets.MenuItem`.

        """
        item: Dict[str, Any] = {
            "text": node.title,
            "url": None,
            "icon": None,
            "items": [],
        }
        if node.page:
            item["url"] = node.page.get_absolute_url()
        return item

    def build(self, items: List[Dict[str, Any]], node: TreeNode) -> None:
        """
        Build a :py:class:`wildewdigets.MenuItem` compatible
        dict representing ``node``, and append it to ``items``.

        if ``node`` has children, recurse into those children, building out
        our submenus.

        Args:
            items: the current list of ``MenuItem`` compatible dicts for the
                current level of the menu
            node: the current node in our page tree

        """
        item = self.build_item(node)
        if node.children:
            for child in node.children:
                self.build(item["items"], child)
        items.append(item)

    def run(self, version: "Version") -> List[Dict[str, Any]]:
        """
        Parse the :py:func:`Version.page_tree` and return a struct that works
        with
        :py:class:`sphinx_hosting.wildewidgets.SphinxPageGlobalTableOfContentsMenu.parse_obj`

        The returned struct should look something like this::

            [
                {'text': 'foo'},
                {'text': 'bar', 'url': '/foo', 'icon': None}
                {'text': 'bar', 'url': '/foo', 'icon': None, items: [{'text': 'blah' ...} ...]}
                ...
            ]

        Args:
            version: the version whose global table of contents we are parsing

        Returns:
            A list of dicts representing the global menu structure

        """  # noqa: E501
        self.sphinx_tree = version.page_tree
        items: List[Dict[str, Any]] = []
        items.append(self.build_item(self.sphinx_tree.head))
        for child in self.sphinx_tree.head.children:
            self.build(items, child)
        return items


class SphinxGlobalTOCHTMLProcessor:
    """
    **Usage**: ``SphinxGlobalTOCHTMLProcessor().run(version, globaltoc_html)```

    This importer is used to parse the ``globaltoc`` key in JSON output of
    Sphinx pages built with the `sphinxcontrib-jsonglobaltoc
    <https://github.com/caltechads/sphinxcontrib-jsonglobaltoc>`_ extension.

    Sphinx uses your ``.. toctree:`` declarations in your ``.rst`` files to
    build site navigation for your document tree, and
    ``sphinxcontrib-jsonglobaltoc`` saves the Sphinx HTML produced by those
    ``..toctree`` as the ``globaltoc`` key in the `.fjson` output.

    Note:
        Sphinx ``.. toctree:`` are ad-hoc -- they're up to how the author wants
        to organize their content, and may not reflect how files are filled out
        in the filesystem.

    """

    def __init__(self, max_level: int = 2) -> None:
        #: Generate at most this many levels of menus and submenus
        self.max_level: int = max_level

    def fix_href(self, href: str) -> str:
        p = urlparse(unquote(href))
        url = reverse(
            "sphinx_hosting:sphinxpage--detail",
            kwargs={
                "project_slug": self.version.project.machine_name,  # type: ignore[attr-defined]
                "version": self.version.version,
                "path": p.path.strip("/"),
            },
        )
        if p.fragment:
            url += f"#{p.fragment}"
        return url

    def parse_ul(self, html: HtmlElement, level: int = 1) -> List[Dict[str, Any]]:
        """
        Process ``html``, an ``lxml`` parsed set of elements representing the contents
        of a ``<ul>`` from a Sphinx table of contents and return a list of
        :py:class:`sphinx_hosting.wildewidgets.MenuItem` objects.

        Any ``href`` in links found will be converted to its full
        ``django-sphinx-hosting`` path.

        If we find another ``<ul>`` inside ``html``, process it by passing its
        contents to :py:meth:`parse_ul` again, incrementing the menu level.

        If ``level`` is greater than :py:attr:`max_level`, return an empty list,
        stopping our recursion.

        Args:
            html: the list of elements that are the contents of the parent ``<ul>``

        Keyword Args:
            level: the current menu level

        Returns:
            The ``<ul>`` contents as a list of dicts

        """
        items: List[Dict[str, Any]] = []
        if level <= self.max_level:
            for li in html.iterchildren():
                item: Dict[str, Any] = {
                    "text": "placeholder",
                    "url": None,
                    "icon": None,
                    "items": [],
                }
                for elem in li:
                    if elem.tag == "a":
                        item["text"] = elem.text_content()
                        item["url"] = self.fix_href(elem.attrib["href"])
                    if elem.tag == "ul":
                        item["items"].extend(self.parse_ul(elem, level=level + 1))
                items.append(item)
        return items

    def parse_globaltoc(self, html: HtmlElement) -> List[Dict[str, Any]]:
        """
        Parse our global table of contents HTML blob and return a list of
        :py:class:`sphinx_hosting.wildewidgets.MenuItem` objects.

        Add a first node that points to the root doc, also.  The root doc can't
        add itself to its ``toctree`` blocks, so we need to do it ourselves.

        How our mapping works:

        * Multiple top level ``<ul>`` tags separated by ``<p class="caption">``
          tags will be merged into a single list.
        * ``<p class="caption ...">CONTENTS</p>`` becomes ``{'text': 'CONTENTS'}```
        * Any ``href`` will be converted to its full ``django-sphinx-hosting`` path

        Args:
            version: the version whose global table of contents we are parsing
            html: the lxml parsed HTML of the global table of contents from
                Sphinx

        """
        root_url = reverse(
            "sphinx_hosting:sphinxpage--detail",
            kwargs={
                "project_slug": self.version.project.machine_name,  # type: ignore[attr-defined]
                "version": self.version.version,
                "path": self.version.head.relative_path,  # type: ignore[attr-defined]
            },
        )
        items: List[Dict[str, Any]] = [
            {"text": "Home", "url": root_url, "icon": None, "items": []}
        ]
        for elem in html.iterchildren():
            if elem.tag == "p" and "caption" in elem.classes:
                # Captions only appear at the top level, even if you assign
                # captions in your toctree declaration in your sub levels with
                # :caption:, so we only process them here.
                items.append({"text": elem.text_content()})
            if elem.tag == "ul":
                items.extend(self.parse_ul(elem))
        return items

    def run(self, version: "Version", verbose: bool = False) -> List[Dict[str, Any]]:
        """
        Parse the global table of contents found as
        ``version.head.orig_global_toc`` into a data struct suitable for use
        with
        :py:class:`sphinx_hosting.wildewidgets.SphinxPageGlobalTableOfContentsMenu.parse_obj`
        and return it.

        How our mapping works:

        * Multiple top level ``<ul>`` tags separated by ``<p class="caption">``
          tags will be merged into a single list.
        * ``<p class="caption ...">CONTENTS</p>`` becomes ``{'text': 'CONTENTS'}```
        * Any ``href`` for links found will be converted to its full
          ``django-sphinx-hosting`` path

        The returned struct should look something like this::

            [
                {'text': 'foo'},
                {'text': 'bar', 'url': '/project/version/foo', 'icon': None}
                {'text': 'bar', 'url': '/project/version/bar', 'icon': None, items: [{'text': 'blah' ...} ...]}
                ...
            ]

        Args:
            version: the version whose global table of contents we are parsing

        Keyword Args:
            verbose: if ``True``, pretty print the HTML of the globaltoc

        Returns:
            A list of dicts representing the global menu structure

        """  # noqa: E501
        self.version = version
        if self.version.head.orig_global_toc:  # type: ignore[attr-defined]
            # This is really only necessary to get the pretty printer below to
            # work properly.  If the \n chars are not removed, lxml won't indent
            # properly
            global_toc_html = re.sub(r"\n", "", self.version.head.orig_global_toc)  # type: ignore[attr-defined]
            html = lxml.html.fromstring(global_toc_html)
            if verbose:
                print(
                    etree.tostring(  # pylint: disable=c-extension-no-member
                        html, method="xml", encoding="unicode", pretty_print=True
                    )
                )
            return self.parse_globaltoc(html)
        return []


# --------------------------
# FileField upload functions
# --------------------------


def sphinx_image_upload_to(instance: "SphinxImage", filename: str) -> str:
    """
    Set the upload path within our ``MEDIA_ROOT`` for any images used by our Sphinx
    documentation to be::

        {project machine_name}/{version}/images/{image basename}

    Args:
        instance: the :py:class:`SphinxImage` object
        filename: the original path to the file

    Returns:
        The properly formatted path to the file

    """
    path = (
        Path(instance.version.project.machine_name)  # type: ignore[attr-defined]
        / Path(instance.version.version)  # type: ignore[attr-defined]
        / "images"
    )
    path = path / Path(filename).name
    return str(path)


def sphinx_document_upload_to(instance: "SphinxDocument", filename: str) -> str:
    """
    Set the upload path within our ``MEDIA_ROOT`` for any images used by our Sphinx
    documentation to be::

        {project machine_name}/{version}/document/{image basename}

    Args:
        instance: the :py:class:`SphinxDocument` object
        filename: the original path to the file

    Returns:
        The properly formatted path to the file

    """
    path = (
        Path(instance.version.project.machine_name)  # type: ignore[attr-defined]
        / Path(instance.version.version)  # type: ignore[attr-defined]
        / "documents"
    )
    path = path / Path(filename).name
    return str(path)


# --------------------------
# Managers
# --------------------------


class ClassifierManager(models.Manager):
    """
    Manager for :py:class:`Classifier` models.
    """

    def tree(self) -> Dict[str, ClassifierNode]:
        """
        Given our classifiers, which are ``::`` separated lists of terms
        like::

            Section :: Subsection :: Name
            Section :: Subsection :: Name2
            Section :: Subsection :: Name3
            Section :: Subsection

        Return a tree-like data structure that looks like::

            {
                'Section': ClassifierNode(
                    title='Section'
                    items={
                        'Subsection': ClassifierNode(
                            title='Subsection',
                            classifier=Classifier(name="Section :: Subsection"),
                            items: {
                                'Name': ClassifierNode(
                                    title='Name',
                                    classifier=Classifier(
                                        name='Section :: Subsection :: Name'
                                    )
                                ),
                                ...
                            }
                        )
                    }
                )
            }
        """
        nodes: Dict[str, ClassifierNode] = {}
        current: ClassifierNode | None = None
        for classifier in self.get_queryset().all():
            parts = classifier.name.split(" :: ")
            if parts[0] not in nodes:
                nodes[parts[0]] = ClassifierNode(title=parts[0])
            current = nodes[parts[0]]
            if len(parts) > 1:
                for part in parts[1:]:
                    if part not in current.items:
                        current.items[part] = ClassifierNode(title=part)
                    current = current.items[part]
            current.classifier = classifier
        return nodes


# --------------------------
# Models
# --------------------------


class Classifier(ViewSetMixin, models.Model):
    """
    A :py:class:`Project` can be tagged with one or more :py:class:`Classifier`
    tags.  This allows you to group projects by ecosystem, or type, for example.

    Use `PyPI <www.pypi.org>`_ classifiers as an example of how to use a single
    field for classifying across many dimensions.

    Examples::

        Ecosystem :: CMS
        Language :: Python
        Owner :: DevOps :: AWS
    """

    objects = ClassifierManager()

    name: F = models.CharField(
        "Classifier Name",
        help_text=_(
            'The classifier spec for this classifier, e.g. "Language :: Python"'
        ),
        max_length=255,
        unique=True,
    )

    def save(self, *args, **kwargs) -> None:
        """
        Overrides :py:meth:`django.db.models.Model.save`.

        Override save to create any missing classifiers in our chain.  For example,
        if we want to create this classifier::

            Foo :: Bar :: Baz

        But ``Foo :: Bar`` does not yet exist in the database, create that before
        creating ``Foo :: Bar :: Baz``.  We do this so that when we filter our projects
        by classifier, we can filter by ``Foo :: Bar`` and ``Foo :: Bar :: Baz``.

        Args:
            args: the arguments to pass to :py:meth:`django.db.models.Model.save`
            kwargs: the keyword arguments to pass to
                :py:meth:`django.db.models.Model.save`

        """
        parts = [p.strip() for p in self.name.split("::")]
        if len(parts) > 2:  # noqa: PLR2004
            name = parts[0]
            for part in parts[1:-1]:
                name = f"{name} :: {part}"
                if not Classifier.objects.filter(name=name).exists():
                    new_classifiier = Classifier(name=name)
                    new_classifiier.save(
                        using=kwargs.get("using", settings.DEFAULT_DB_ALIAS)
                    )
        # Rejoin our parts to ensure we always get a classifier that looks like
        # "part :: part :: part" instead of "part:: part::part"
        self.name = " :: ".join(parts)
        super().save(*args, **kwargs)

    def __str__(self) -> str:
        return str(self.name)

    class Meta:
        verbose_name = _("classifier")
        verbose_name_plural = _("classifiers")


class ProjectPermissionGroup(ViewSetMixin, TimeStampedModel, models.Model):
    """
    A :py:class:`Project` can be assigned to one or more
    :py:class:`ProjectPermissionGroup` groups.  This restricts viewing of the
    project to users which belong to those groups.

    Examples:
        .. code-block:: text

            Ecosystem :: CMS
            Language :: Python
            Owner :: DevOps :: AWS

    """

    name: F = models.CharField(
        "Permission Group Name",
        help_text=_("The name for this permission group"),
        max_length=100,
        unique=True,
    )
    description: F = models.CharField(
        "Brief Description",
        max_length=256,
        null=True,
        blank=True,
        help_text=_("A brief description of this permission group"),
        validators=[NoHTMLValidator()],
    )

    users: M2M = models.ManyToManyField(
        settings.AUTH_USER_MODEL, related_name="project_permission_groups"
    )

    class Meta:
        verbose_name = _("project permission group")
        verbose_name_plural = _("project permission groups")


class Project(ViewSetMixin, TimeStampedModel, models.Model):
    """
    A Project is what a set of Sphinx docs describes: an application, a library, etc.

    Projects have versions (:py:class:`Version`) and versions have Sphinx pages
    (:py:class:`SphinxPage`).
    """

    title: F = models.CharField(
        "Project Name", help_text=_("The human name for this project"), max_length=100
    )
    description: F = models.CharField(
        "Brief Description",
        max_length=256,
        null=True,
        blank=True,
        help_text=_("A brief description of this project"),
        validators=[NoHTMLValidator()],
    )
    machine_name: F = MachineNameField(
        "Machine Name",
        unique=True,
        help_text=_(
            """Must be unique.  Set this to the slugified value of "project" in """
            """Sphinx's. conf.py"""
        ),
    )

    latest_version: FK = models.ForeignKey(
        "Version",
        help_text=_(
            "The latest version of this project.  "
            "This is the version that will be shown when you click "
            '"Read Docs" on the project page.'
        ),
        null=True,
        blank=True,
        related_name="+",
        on_delete=models.SET_NULL,
    )

    permission_groups: M2M = models.ManyToManyField(
        ProjectPermissionGroup,
        related_name="projects",
    )
    classifiers: M2M = models.ManyToManyField(
        Classifier,
        related_name="projects",
    )

    def __str__(self) -> str:  # pylint: disable=invalid-str-returned
        return self.title

    def get_absolute_url(self) -> str:
        return reverse("sphinx_hosting:project--detail", args=[self.machine_name])

    def get_update_url(self) -> str:
        return reverse("sphinx_hosting:project--update", args=[self.machine_name])

    def get_latest_version_url(self) -> str | None:
        if self.latest_version:
            return reverse(
                "sphinx_hosting:sphinxpage--detail",
                args=[
                    self.machine_name,
                    self.latest_version.version,  # type: ignore[attr-defined]
                    self.latest_version.head.relative_path,  # type: ignore[attr-defined]
                ],
            )
        return None

    class Meta:
        verbose_name = _("project")
        verbose_name_plural = _("projects")


class Version(TimeStampedModel, models.Model):
    """
    A ``Version`` is a specific version of a :py:class:`Project`. Versions own
    :py:class:`SphinxPage` objects.
    """

    project: FK = models.ForeignKey(
        Project,
        on_delete=models.CASCADE,
        related_name="versions",
        help_text=_("The Project to which this Version belongs"),
    )
    version: F = models.CharField(
        "Version",
        max_length=64,
        null=False,
        help_text=_("The version number for this release of the Project"),
    )
    sphinx_version: F = models.CharField(
        "Sphinx Version",
        max_length=64,
        null=True,
        blank=True,
        default=None,
        help_text=_("The version of Sphinx used to create this documentation set"),
    )
    archived: F = models.BooleanField(
        "Archived?",
        default=False,
        help_text=_("Whether this version should be excluded from search indexes"),
    )

    head: FK = models.OneToOneField(
        "SphinxPage",
        on_delete=models.SET_NULL,
        null=True,
        related_name="+",  # disable our related_name for this one
        help_text=_(
            "The top page of the documentation set for this version of our project"
        ),
    )

    def __str__(self) -> str:
        return f"{self.project.title}-{self.version}"  # type: ignore[attr-defined]

    @property
    def is_latest(self) -> bool:
        return self == self.project.latest_version  # type: ignore[attr-defined]

    @property
    def page_tree(self) -> SphinxPageTree:
        """
        Return the page hierarchy for the set of :py:class:`SphinxPage` pages in
        this version.

        The page hierarchy is build by traversing the pages in the set, starting
        with :py:attr:`head`.

        Returns:
            The page hierarchy for this version.

        """
        return SphinxPageTree(self)

    def mark_searchable_pages(self) -> None:
        """
        Set the :py:attr:`SphinxPage.searchable` flag on
        the searchable pages in this version.

        Searchable pages are ones that:

        * Are not in :py:attr:`SphinxPage.SPECIAL_PAGES`
        * Do not have a part of their relative path that starts with ``_``.

        Go through the pages in this version, and set
        :py:attr:`SphinxPage.searchable` to ``True`` for all those which meet the above
        requirements, ``False`` otherwise.
        """
        ignored_paths = list(SphinxPage.SPECIAL_PAGES.keys())
        for page in self.pages.all():
            if (
                page.relative_path not in ignored_paths
                and not page.relative_path.startswith("_")
                and "/_" not in page.relative_path
            ):
                page.searchable = True
            else:
                page.searchable = False
            page.save()

    @cached_property
    def globaltoc(self) -> Dict[str, List[Dict[str, Any]]]:
        """
        Build a struct that looks like this::

            {
                items: [
                    {'text': 'foo'},
                    {'text': 'bar', 'url': '/foo', 'icon': None}
                    {'text': 'bar', 'url': '/foo', 'icon': None, items: [{'text': 'blah' ...} ...]}
                    ...
                ]
            }

        suitable for constructing a
        :py:class:`sphinx_hosting.wildewidgets.SphinxPageGlobalTableOfContentsMenu`
        for this :py:class:`Version`.

        """  # noqa: E501
        items = SphinxGlobalTOCHTMLProcessor(max_level=MAX_GLOBAL_TOC_TREE_DEPTH).run(
            self
        )
        if not items:
            items = SphinxPageTreeProcessor().run(self)
        return {"items": items}

    def purge_cached_globaltoc(self) -> None:
        """
        Purge the cached output from our :py:meth:`globaltoc` property.
        """
        with suppress(AttributeError):
            # If we get AttributeError, this means self.globaltoc hasn't been
            # accessed yet
            del self.globaltoc

    def get_absolute_url(self) -> str:
        return reverse(
            "sphinx_hosting:version--detail",
            args=[self.project.machine_name, self.version],  # type: ignore[attr-defined]
        )

    def save(self, *args, **kwargs):
        """
        Overriding :py:meth:`django.db.models.Model.save` here so that we can
        purge our cached global table of contents.
        """
        super().save(*args, **kwargs)
        self.purge_cached_globaltoc()


class SphinxPage(TimeStampedModel, models.Model):
    """
    A ``SphinxPage`` is a single page of a set of Sphinx documentation.
    ``SphinxPage`` objects are owned by :py:class:`Version` objects, which are
    in turn owned by :py:class:`Project` objects.
    """

    #: This is a mapping between filename and title that identifies the
    #: special pages that Sphinx produces on its own and gives them
    #: reasonable titles.  These pages have no ``title`` key in their
    #: json data, but ``title`` is required for pages
    SPECIAL_PAGES: Final[Dict[str, str]] = {
        "genindex": "General Index",
        "py-modindex": "Module Index",
        "np-modindex": "Module Index",
        "search": "Search",
        "_modules/index": "Module code",
    }

    version: FK = models.ForeignKey(
        Version,
        on_delete=models.CASCADE,
        related_name="pages",
        help_text=_("The Version to which this page belongs"),
    )
    relative_path: F = models.CharField(
        "Relative page path",
        help_text=_("The path to the page under our top slug"),
        max_length=255,
    )
    content: F = models.TextField(
        "Content", help_text=_("The full JSON payload for the page")
    )
    title: F = models.CharField(
        "Title",
        max_length=255,
        help_text=_("Just the title for the page, extracted from the page JSON"),
    )
    orig_body: F = models.TextField(
        "Body (Original)",
        blank=True,
        help_text=_(
            "The original body for the page, extracted from the page JSON. Some pages "
            "have no body. We save this here in case we need to reprocess the body at "
            "some later date."
        ),
    )
    body: F = models.TextField(
        "Body",
        blank=True,
        help_text=_(
            "The body for the page, extracted from the page JSON, and modified to "
            "suit us. Some pages have no body.  The body is actually stored as a "
            "Django template."
        ),
    )
    orig_local_toc: F = models.TextField(
        "Local Table of Contents (original)",
        blank=True,
        null=True,
        default=None,
        help_text=_(
            "The original table of contents for headings in this page."
            "We save this here in case we need to reprocess the table of contents "
            "at some later date."
        ),
    )
    local_toc: F = models.TextField(
        "Local Table of Contents",
        blank=True,
        null=True,
        default=None,
        help_text=_(
            "Table of Contents for headings in this page, modified to work in "
            "our templates"
        ),
    )
    orig_global_toc: F = models.TextField(
        "Global Table of Contents (original)",
        blank=True,
        null=True,
        default=None,
        help_text=_(
            "The original global table of contents HTML attached to this page, if any. "
            ' This will only be present if you had "sphinxcontrib-jsonglobaltoc"'
            'installed in your "extensions" in the Sphinx conf.py'
        ),
    )
    searchable: F = models.BooleanField(
        "Searchable",
        default=False,
        help_text=_("Should this page be included in the search index?"),
    )

    parent: FK = models.ForeignKey(
        "SphinxPage",
        on_delete=models.CASCADE,
        null=True,
        related_name="children",
        help_text=_("The parent page of this page"),
    )

    # This has to be a ForeignKey here and not a OneToOneField becuase
    # more than one page can have the same next_page.
    next_page: FK = models.ForeignKey(
        "SphinxPage",
        on_delete=models.CASCADE,
        null=True,
        related_name="previous_page",
        help_text=_("The next page in the documentation set"),
    )

    def __str__(self) -> str:
        return (
            f"{self.version.project.title}-{self.version.version}: {self.relative_path}"  # type: ignore[attr-defined]
        )

    def get_absolute_url(self) -> str:
        return reverse(
            "sphinx_hosting:sphinxpage--detail",
            args=[
                self.version.project.machine_name,  # type: ignore[attr-defined]
                self.version.version,  # type: ignore[attr-defined]
                self.relative_path,
            ],
        )

    def get_permalink(self) -> str:
        """
        Return the permalink for this page.  This is the URL for the page
        with the version number replaced with "latest".

        Returns:
            The permalink for this page.

        """
        return reverse(
            "sphinx_hosting:sphinxpage--detail",
            args=[self.version.project.machine_name, "latest", self.relative_path],  # type: ignore[attr-defined]
        )

    class Meta:
        verbose_name = _("sphinx page")
        verbose_name_plural = _("sphinx pages")
        unique_together = ("version", "relative_path")


class SphinxImage(TimeStampedModel, models.Model):
    """
    A ``SphinxImage`` is an image file referenced in a Sphinx document.  When
    importing documenation packages, we extract all images from the package,
    upload them into Django storage and update the Sphinx HTML in
    :py:attr:`SphinxPage.body` to reference the URL for the uploaded image
    instead of its original url.
    """

    version: FK = models.ForeignKey(
        Version,
        on_delete=models.CASCADE,
        related_name="images",
        help_text=_(
            "The version of our project documentation with which this image is "
            "associated"
        ),
    )
    orig_path: F = models.CharField(
        _("Original Path"),
        max_length=256,
        help_text=_(
            "The original path to this file in the Sphinx documentation package"
        ),
    )
    file: F = models.FileField(
        _("An image file"),
        upload_to=sphinx_image_upload_to,
        help_text=_("The actual image file"),
    )

    class Meta:
        verbose_name = _("sphinx image")
        verbose_name_plural = _("sphinx images")
        unique_together = ("version", "orig_path")


class SphinxDocument(TimeStampedModel, models.Model):
    """
    A ``SphinxDocument`` is an downloadable document file referenced in a Sphinx
    document.  When importing documenation packages, we extract all such
    document from the package, upload them into Django storage and update the
    Sphinx HTML in :py:attr:`SphinxPage.body` to reference the URL for the
    uploaded document instead of its original url.
    """

    version: FK = models.ForeignKey(
        Version,
        on_delete=models.CASCADE,
        related_name="documents",
        help_text=_(
            "The version of our project documentation with which this document is "
            "associated"
        ),
    )
    orig_path: F = models.CharField(
        _("Original Path"),
        max_length=256,
        help_text=_(
            "The original path to this file in the Sphinx documentation package"
        ),
    )
    file: F = models.FileField(
        _("An image file"),
        upload_to=sphinx_document_upload_to,
        help_text=_("The actual document file"),
    )

    class Meta:
        verbose_name = _("sphinx document")
        verbose_name_plural = _("sphinx documents")
        unique_together = ("version", "orig_path")


class ProjectRelatedLink(TimeStampedModel, models.Model):
    """
    A ``ProjectRelatedLink`` is a link to an external resource that is related
    to a :py:class:`Project`.
    """

    title: F = models.CharField(
        "Link Title", max_length=100, help_text=_("The title for this link")
    )
    uri: F = models.URLField(
        "Link URL", max_length=256, help_text=_("The URL for this link")
    )
    project: FK = models.ForeignKey(
        Project,
        on_delete=models.CASCADE,
        related_name="related_links",
        help_text=_("The project to which this link is related"),
    )

    def get_update_url(self) -> str:
        return reverse(
            "sphinx_hosting:projectrelatedlink--update", kwargs={"pk": self.pk}
        )

    def get_delete_url(self) -> str:
        return reverse(
            "sphinx_hosting:projectrelatedlink--delete", kwargs={"pk": self.pk}
        )

    class Meta:
        verbose_name = _("project related link")
        verbose_name_plural = _("project related links")
        unique_together = ("project", "uri")
        ordering = ("title",)



from typing import Type  # noqa: UP035

from django.contrib.auth.models import AbstractUser, Group, UserManager
from django.db.models import Model
from django.db.models.signals import post_save
from django.dispatch import receiver
from rest_framework.authtoken.models import Token


class User(AbstractUser):
    class Meta:
        verbose_name = "user"
        verbose_name_plural = "users"


# API-only users
# -----------------------


class APIUserManager(UserManager):
    def get_queryset(self):
        return User.objects.filter(groups__name="API Users")


class APIUser(User):
    """
    These are API specific users -- they get no access to the django admin
    console, and they have API tokens.
    """

    objects = APIUserManager()

    def clean(self):
        if not self.username.startswith("api-"):
            self.username = "api-" + self.username

    class Meta:
        proxy = True
        verbose_name = "API User"
        verbose_name_plural = "API Users"


@receiver(post_save, sender=APIUser)
def finalize_api_user(sender: Type[Model], instance: Model, created: bool, **kwargs):  # noqa: ARG001
    if created:
        g = Group.objects.get(name="API Users")
        g.user_set.add(instance)  # type: ignore[attr-defined]
        Token.objects.get_or_create(user=instance)



# Generated by Django 5.1.4 on 2024-12-17 17:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sphinxhostingcore', '0016_project_last_version_alter'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sphinxpage',
            name='body',
            field=models.TextField(blank=True, help_text='The body for the page, extracted from the page JSON, and modified to suit us. Some pages have no body.  The body is actually stored as a Django template.', verbose_name='Body'),
        ),
        migrations.AlterField(
            model_name='sphinxpage',
            name='orig_global_toc',
            field=models.TextField(blank=True, default=None, help_text='The original global table of contents HTML attached to this page, if any.  This will only be present if you had "sphinxcontrib-jsonglobaltoc"installed in your "extensions" in the Sphinx conf.py', null=True, verbose_name='Global Table of Contents (original)'),
        ),
    ]



from typing import Any, Final

from django.db.models import Model
from rest_framework import serializers
from rest_framework.relations import RelatedField

from sphinx_hosting.models import (
    Classifier,
    Project,
    ProjectRelatedLink,
    SphinxImage,
    SphinxPage,
    Version,
)


class ClassifierSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model: type[Model] = Classifier
        fields: tuple[str, ...] = ("url", "id", "name")
        extra_kwargs: Final[dict[str, Any]] = {
            "url": {"view_name": "sphinx_hosting_api:classifier-detail"},
        }


class ProjectSerializer(serializers.HyperlinkedModelSerializer):
    versions: RelatedField = serializers.HyperlinkedRelatedField(
        many=True, read_only=True, view_name="sphinx_hosting_api:version-detail"
    )
    latest_version: RelatedField = serializers.HyperlinkedRelatedField(
        many=False, read_only=True, view_name="sphinx_hosting_api:version-detail"
    )
    classifiers: serializers.Serializer = ClassifierSerializer(many=True)
    related_links: RelatedField = serializers.HyperlinkedRelatedField(
        many=True,
        read_only=True,
        view_name="sphinx_hosting_api:projectrelatedlink-detail",
    )

    class Meta:
        model: type[Model] = Project
        read_only_fields: tuple[str, ...] = ("machine_name",)
        fields: tuple[str, ...] = (
            "url",
            "id",
            "title",
            "machine_name",
            "description",
            "related_links",
            "classifiers",
            "latest_version",
            "versions",
        )
        extra_kwargs: Final[dict[str, Any]] = {
            "url": {"view_name": "sphinx_hosting_api:project-detail"},
        }


class ProjectRelatedLinkSerializer(serializers.HyperlinkedModelSerializer):
    project: RelatedField = serializers.HyperlinkedRelatedField(
        queryset=Project.objects.all(), view_name="sphinx_hosting_api:project-detail"
    )

    class Meta:
        model: type[Model] = ProjectRelatedLink
        fields: tuple[str, ...] = (
            "url",
            "id",
            "title",
            "uri",
            "project",
        )
        extra_kwargs: Final[dict[str, Any]] = {
            "url": {"view_name": "sphinx_hosting_api:projectrelatedlink-detail"},
        }


class VersionSerializer(serializers.HyperlinkedModelSerializer):
    project: RelatedField = serializers.HyperlinkedRelatedField(
        read_only=True, view_name="sphinx_hosting_api:project-detail"
    )
    head: RelatedField = serializers.HyperlinkedRelatedField(
        read_only=True, view_name="sphinx_hosting_api:sphinxpage-detail"
    )
    pages: RelatedField = serializers.HyperlinkedRelatedField(
        many=True, read_only=True, view_name="sphinx_hosting_api:sphinxpage-detail"
    )
    images: RelatedField = serializers.HyperlinkedRelatedField(
        many=True, read_only=True, view_name="sphinx_hosting_api:sphinximage-detail"
    )

    class Meta:
        model: type[Model] = Version
        fields: tuple[str, ...] = (
            "url",
            "id",
            "project",
            "version",
            "sphinx_version",
            "archived",
            "head",
            "pages",
            "images",
        )
        extra_kwarg: Final[dict[str, Any]] = {
            "url": {"view_name": "sphinx_hosting_api:version-detail"},
        }


class VersionUploadSerializer(serializers.Serializer):
    """
    The actual work of importing the file is done in
    :py:class:`sphinx_hosting.api.views.VersionUploadView`.  We're defining our
    :py:meth:`create` and :py:meth:`create` here as NOOP functions to make the
    linters happy because they're abstract in :py:class:`serializers.Serializer`.
    """

    file = serializers.FileField()

    def create(self, validated_data: dict[str, Any]) -> None:
        pass

    def update(self, instance, validated_data):
        pass

    class Meta:
        fields = ("file",)


class SphinxPageSerializer(serializers.HyperlinkedModelSerializer):
    version: RelatedField = serializers.HyperlinkedRelatedField(
        read_only=True, view_name="sphinx_hosting_api:version-detail"
    )
    parent: RelatedField = serializers.HyperlinkedRelatedField(
        read_only=True, view_name="sphinx_hosting_api:sphinxpage-detail"
    )
    previous_page: RelatedField = serializers.HyperlinkedRelatedField(
        many=True, read_only=True, view_name="sphinx_hosting_api:sphinxpage-detail"
    )
    next_page: RelatedField = serializers.HyperlinkedRelatedField(
        read_only=True, view_name="sphinx_hosting_api:sphinxpage-detail"
    )

    class Meta:
        model: type[Model] = SphinxPage
        fields: tuple[str, ...] = (
            "url",
            "id",
            "version",
            "title",
            "relative_path",
            "content",
            "orig_body",
            "body",
            "orig_local_toc",
            "local_toc",
            "orig_global_toc",
            "searchable",
            "parent",
            "next_page",
            "previous_page",
        )
        extra_kwargs: Final[dict[str, Any]] = {
            "url": {"view_name": "sphinx_hosting_api:sphinxpage-detail"},
        }


class SphinxImageSerializer(serializers.HyperlinkedModelSerializer):
    version: RelatedField = serializers.HyperlinkedRelatedField(
        many=True, read_only=True, view_name="sphinx_hosting_api:version-detail"
    )

    class Meta:
        model: type[Model] = SphinxImage
        fields: tuple[str, ...] = (
            "url",
            "id",
            "version",
            "orig_path",
        )
        extra_kwargs: Final[dict[str, Any]] = {
            "url": {"view_name": "sphinx_hosting_api:sphinximage-detail"},
        }



# Generated by Django 4.1.7 on 2023-03-27 18:19

import itertools
from typing import ClassVar

from django.contrib.auth.management import create_permissions
from django.db import migrations


def apply_migration(apps, schema_editor):
    """
    Create the default auth groups and assign their permissions.

    Args:
        apps: The historical app registry for this migration state.
        schema_editor: The active schema editor for the migration run.

    Side Effects:
        Creates ``auth.Group`` rows and assigns matching permissions.

    """
    # Force the creation of the permission objects we'll need
    for app_config in apps.get_app_configs():
        app_config.models_module = True
        create_permissions(app_config, verbosity=0)
        app_config.models_module = None
    db_alias = schema_editor.connection.alias
    Group = apps.get_model("auth", "Group")
    Permission = apps.get_model("auth", "Permission")
    # Create our groups
    Group.objects.using(db_alias).bulk_create(
        [
            Group(name="Administrators"),
            Group(name="Classifier Managers"),
            Group(name="PermissionGroup Managers"),
            Group(name="Project Managers"),
            Group(name="Version Managers"),
            Group(name="Editors"),
        ]
    )
    # Now assign permissions
    group = Group.objects.using(db_alias).get(name="Classifier Managers")
    classifier_permissions = Permission.objects.filter(codename__endswith="classifier")
    group.permissions.set(classifier_permissions)

    group = Group.objects.using(db_alias).get(name="Project Managers")
    project_permissions = Permission.objects.filter(codename__endswith="project")
    group.permissions.set(project_permissions)

    group = Group.objects.using(db_alias).get(name="PermissionGroup Managers")
    permissiongroup_permissions = Permission.objects.filter(
        codename__endswith="permissiongroup"
    )
    group.permissions.set(permissiongroup_permissions)

    group = Group.objects.using(db_alias).get(name="Version Managers")
    version_permissions = Permission.objects.filter(codename__endswith="version")
    group.permissions.set(version_permissions)

    group = Group.objects.using(db_alias).get(name="Administrators")
    group.permissions.set(
        itertools.chain(
            classifier_permissions,
            permissiongroup_permissions,
            project_permissions,
            version_permissions,
        )
    )
    group = Group.objects.using(db_alias).get(name="Editors")
    group.permissions.set(itertools.chain(project_permissions, version_permissions))


def revert_migration(apps, schema_editor):
    """
    Remove the auth groups created by :func:`apply_migration`.

    Args:
        apps: The historical app registry for this migration state.
        schema_editor: The active schema editor for the migration run.

    Side Effects:
        Deletes the groups created by this migration.

    """
    del schema_editor
    Group = apps.get_model("auth", "Group")
    Group.objects.filter(
        name__in=[
            "Classifier Managers",
            "PermissionGroup Managers",
            "Project Managers",
            "Version Managers",
            "Editors",
            "Administrators",
        ]
    ).delete()


class Migration(migrations.Migration):
    """
    Create default ``django-sphinx-hosting`` auth groups and permissions.
    """

    #: Migrations that must run before this migration.
    dependencies: ClassVar[list[tuple[str, str]]] = [
        ("sphinxhostingcore", "0009_SphinxPage_searchable"),
    ]

    #: Operations applied by this migration.
    operations: ClassVar[list[migrations.RunPython]] = [
        migrations.RunPython(apply_migration, revert_migration)
    ]



# Generated by Django 5.0.7 on 2024-07-10 22:13

from django.db import migrations


def set_latest_version(apps, schema_editor):
    """
    Set the :py:attr:`sphinx_hosting.models.Project.latest_version` field for all
    projects to the latest version using the same logic as the @property
    latest_version in :py:class:`sphinx_hosting.models.Project` for versions < 1.4.2
    """
    Project = apps.get_model("sphinxhostingcore", "Project")
    for project in Project.objects.all():
        latest_version = project.versions.order_by('-modified').first()
        project.latest_version = latest_version
        project.save()


def unset_latest_version(apps, schema_editor):
    """
    Set the :py:attr:`sphinx_hosting.models.Project.latest_version` field to
    None for all projects.   We're undoing the work of the ``set_latest_version
    function``.
    """
    Project = apps.get_model("sphinxhostingcore", "Project")
    for project in Project.objects.all():
        project.latest_version = None
        project.save()


class Migration(migrations.Migration):

    dependencies = [
        ('sphinxhostingcore', '0014_latest_version_as_field'),
    ]

    operations = [
        migrations.RunPython(set_latest_version, reverse_code=unset_latest_version),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING, Final

from wildewidgets import (
    ActionButtonModelTable,
    BasicModelTable,
    Block,
    CardWidget,
    CrispyFormModalWidget,
    CrispyFormWidget,
    Datagrid,
    FormButton,
    HorizontalLayoutBlock,
    Link,
    ListModelWidget,
    ModalButton,
    RowActionButton,
    RowEditButton,
    RowModelUrlButton,
    ToggleableManyToManyFieldBlock,
    TwoColumnLayout,
    Widget,
    WidgetListLayoutHeader,
)

from ..forms import (
    ProjectCreateForm,
    ProjectRelatedLinkCreateForm,
    ProjectRelatedLinkUpdateForm,
)
from ..models import Classifier, Project, ProjectRelatedLink, Version
from .classifier import ClassifierFilterBlock

if TYPE_CHECKING:
    from django.contrib.auth.models import AbstractUser
    from django.db.models import Model, QuerySet

# ------------------------------------------------------
# Modals
# ------------------------------------------------------


class ProjectCreateModalWidget(CrispyFormModalWidget):
    """
    A modal dialog that holds the
    :py:class:`sphinx_hosting.forms.ProjectCreateForm`.

    Args:
        args: the arguments to pass to the modal widget

    Keyword Args:
        kwargs: the keyword arguments to pass to the modal widget

    """

    #: The ID of the modal dialog.
    modal_id: str = "project__create"
    #: The title of the modal dialog.
    modal_title: str = "Add Project"

    def __init__(self, *args, **kwargs):
        form = ProjectCreateForm()
        super().__init__(*args, form=form, **kwargs)


class ProjectRelatedLinkCreateModalWidget(CrispyFormModalWidget):
    """
    A modal dialog that holds the
    :py:class:`sphinx_hosting.forms.ProjectRelatedLinkForm` for
    creating links.

    Args:
        project: the project to create the link for
        args: the arguments to pass to the modal widget

    Keyword Args:
        kwargs: the keyword arguments to pass to the modal widget

    """

    modal_id: str = "projectrelatedlink__create"
    modal_title: str = "Add Related Link"

    def __init__(self, project: Project, *args, **kwargs):
        form = ProjectRelatedLinkCreateForm(project=project)
        super().__init__(*args, form=form, **kwargs)


class ProjectRelatedLinkUpdateModalWidget(CrispyFormModalWidget):
    """
    A modal dialog that holds the
    :py:class:`sphinx_hosting.forms.ProjectRelatedLinkForm` for
    updating links.

    One of these is created for each
    :py:class:`sphinx_hosting.models.ProjectRelatedLink` object, and lives in
    the :py:class:`sphinx_hosting.widgets.ProjectRelatedLinksWidget`.
    """

    modal_title: str = "Update Related Link"

    def __init__(self, link: ProjectRelatedLink, *args, **kwargs):
        form = ProjectRelatedLinkUpdateForm(instance=link)
        super().__init__(*args, form=form, **kwargs)


# ------------------------------------------------------
# Project related widgets
# ------------------------------------------------------


class ProjectInfoWidget(CardWidget):
    """
    A :py:class:`wildewidgets.CardWidget` containing a Tabler datagrid
    that gives information about a :py:class:`sphinx_hosting.models.Project`.
    """

    title: str = "Project Info"
    icon: str = "info-square"

    def __init__(self, project: Project, **kwargs):
        super().__init__(**kwargs)
        grid = Datagrid()
        grid.add_item(title="Machine Name", content=project.machine_name)
        grid.add_item(
            title="Created", content=project.created.strftime("%Y-%m-%d %H:%M %Z")
        )
        grid.add_item(
            title="Last Modified",
            content=project.modified.strftime("%Y-%m-%d %H:%M %Z"),
        )
        self.set_widget(grid)


class ProjectTableWidget(Block):
    """
    A :py:class:`wildewidgets.CardWidget` that gives our
    :py:class:`ProjectTable` dataTable a nice header with a total book count and
    an "Add Project" button that opens a modal dialog.
    """

    def __init__(self, user: AbstractUser, **kwargs):
        super().__init__(**kwargs)
        self.add_block(self.get_title(user))
        layout = TwoColumnLayout(left_column_width=9)
        table = ProjectTable()
        layout.add_to_left(CardWidget(widget=table))
        layout.add_to_right(
            ClassifierFilterBlock(
                table_name=table.table_name,
                column_number=table.get_column_number("classifiers"),
            )
        )
        self.add_block(layout)

    def get_title(self, user: AbstractUser) -> WidgetListLayoutHeader:  # pylint: disable=arguments-differ
        header = WidgetListLayoutHeader(
            header_text="Projects",
            badge_text=Project.objects.count(),
        )
        if user.has_perm("sphinxhostingcore.add_project"):
            header.add_modal_button(
                text="New Project",
                color="primary",
                target=f"#{ProjectCreateModalWidget.modal_id}",
            )
        return header


class ProjectVersionsTableWidget(CardWidget):
    """
    A :py:class:`wildewidgets.CardWidget` that gives our
    :py:class:`ProjectVersionTable` dataTable a nice header with a total version
    count.
    """

    title: str = "Versions"
    icon: str = "bookmark-star"

    def __init__(self, project_id: int, **kwargs):
        self.project_id = project_id
        super().__init__(
            widget=ProjectVersionTable(project_id=project_id),
            **kwargs,
        )

    def get_title(self) -> WidgetListLayoutHeader:
        return WidgetListLayoutHeader(
            header_text="Versions",
            badge_text=Project.objects.get(pk=self.project_id).versions.count(),
        )


class ProjectClassifierSelectorWidget(ToggleableManyToManyFieldBlock):
    """
    An editable listing of :py:class:`sphinx_hosting.models.Classifier`
    objects for a particular :py:class:`sphinx_hosting.models.Project`.

    It is used in the project edit view.
    """

    model: type[Model] = Project
    field_name: str = "classifiers"
    title: str = "Classifiers"
    icon: str = "collection"


class ProjectClassifierListWidget(ListModelWidget):
    """
    A :py:class:`wildewidgets.ListModelWidget` that renders a list of
    :py:class:`sphinx_hosting.models.Classifier` objects.  It is used in the
    project detail view as a read-only list of classifiers.

    The editable version of this widget is
    :py:class:`ProjectClassifierSelectorWidget`.
    """

    paginate_by: int = 100
    item_label: str = "Classifier"
    title = "Classifiers"
    icon = "collection"

    def get_object_text(self, instance: Classifier) -> str:
        return instance.name

    def get_model_subblock(self, instance: Classifier) -> Block:
        return Block(self.get_object_text(instance), tag="label")


class ProjectDetailWidget(CrispyFormWidget, Widget):
    """
    Renders an update form for a :py:class:`sphinx_hosting.models.Project`.

    Use directly it like so::

        >>> project = Project.objects.get(pk=1)
        >>> form = ProjectUpdateForm(instance=project)
        >>> widget = ProjectDetailWidget(form=form)

    Or you can simply add the form to your view context and
    :py:class:`ProjectDetailWidget` will pick it up automatically.
    """

    title: str = "General Settings"
    name: str = "project-detail__section"
    modifier: str = "general"
    icon: str = "card-checklist"
    css_class: str = CrispyFormWidget.css_class + " p-4"


# ------------------------------------------------------
# ProjectRelatedLink related widgets
# ------------------------------------------------------


class ProjectRelatedLinkListItemWidget(HorizontalLayoutBlock):
    """
    Used by :py:class:`ProjectRelatedLinksWidget` to render a single
    :py:class:`sphinx_hosting.models.ProjectRelatedLink` object in its list.
    """

    css_class: str = "mb-2"

    def __init__(self, object: ProjectRelatedLink):  # noqa: A002
        modal_id = f"projectrelatedlink__update__{object.pk}"
        super().__init__(
            Link(object.title, url=object.uri, title=object.title),
            Block(
                ModalButton(text="Edit", color="azure", target=f"#{modal_id}"),
                FormButton(
                    text="Delete",
                    color="outline-secondary",
                    css_class="d-inline",
                    action=object.get_delete_url(),
                    confirm_text="Are you sure you want to delete this link?",
                ),
            ),
            ProjectRelatedLinkUpdateModalWidget(object, modal_id=modal_id),
        )


class ProjectRelatedLinksWidget(CardWidget):
    """
    A :py:class:`wildewidgets.CardWidget` that allows us to manage the
    :py:class:`sphinx_hosting.models.ProjectRelatedLink` objects for this
    :py:class:`sphinx_hosting.models.Project`.

    It renders a list of :py:class:`ProjectRelatedLinkListItemWidget` objects
    and a "Add Related Link" button that opens a modal dialog.

    This is used on the project update page, and is the editable version of
    :py:class:`ProjectRelatedLinksListWidget`.
    """

    title: str = "Related Links"
    icon: str = "box-arrow-up-right"

    def __init__(self, project: Project, **kwargs):
        self.project = project
        super().__init__(
            widget=ListModelWidget(
                queryset=self.project.related_links,
                model_widget=ProjectRelatedLinkListItemWidget,
            ),
            **kwargs,
        )

    def get_title(self) -> WidgetListLayoutHeader:
        header = WidgetListLayoutHeader(
            header_text="Related Links",
            badge_text=self.project.related_links.count(),
        )
        header.add_modal_button(
            text="Add Related Link",
            color="primary",
            target=f"#{ProjectRelatedLinkCreateModalWidget.modal_id}",
        )
        return header


class ProjectRelatedLinksListWidget(ListModelWidget):
    """
    Renders a list of :py:class:`sphinx_hosting.models.ProjectRelatedLink`
    objects as a list of :py:class:`wildewidgets.Link` objects.

    This is used on the project detail page, and is the read-only version of
    :py:class:`ProjectRelatedLinksWidget`.
    """

    paginate_by: int = 100
    item_label: str = "Related Link"
    title: str = "Related Links"
    icon: str = "external-link"

    def get_model_subblock(self, instance: ProjectRelatedLink) -> Block:
        return Link(instance.title, url=instance.uri)


# ------------------------------------------------------
# Datatables
# ------------------------------------------------------


class LatestVersionButton(RowModelUrlButton):
    attribute: str = "get_latest_version_url"
    text: str = "Read Docs"
    color: str = "orange"

    def is_visible(self, row: Project, user: AbstractUser) -> bool:
        if row.latest_version is None:
            return False
        return super().is_visible(row, user)


class ProjectTable(ActionButtonModelTable):
    """
    Displays a `dataTable <https://datatables.net>`_ of our
    :py:class:`sphinx_hosting.models.Project` instances.

    It's used as a the main widget in by :py:class:`ProjectTableWidget`.
    """

    model: type[Model] = Project

    #: Show this many rows per page
    page_length: int = 25
    #: Set to ``True`` to stripe our table rows
    striped: bool = True
    default_action_button_label: str = "Edit"
    default_action_button_color_class: str = "outline-secondary"
    #: A list of fields that we will list as columns.  These are either fields
    #: on our :py:attr:`model`, or defined as ``render_FIELD_NAME_column`` methods
    #: on this class
    fields: Final[list[str]] = [
        "title",
        "machine_name",
        "classifiers",
        "latest_version",
        "description",
        "latest_version_date",
    ]
    #: A list of names of columns to hide by default.
    hidden: Final[list[str]] = [
        "classifiers",
        "machine_name",
        "latest_version_date",
    ]
    #: A list of names of columns that will will not be sortable
    #: (i.e. clicking on the column header will not sort the table)
    unsortable: Final[list[str]] = [
        "latest_version",
        "latest_version_date",
    ]
    #: A list of names of columns that will will not be searched when doing a
    #: **global** search
    unsearchable: Final[list[str]] = [
        "classifiers",
        "latest_version",
        "latest_version_date",
    ]
    #: A dict of column name to column label.  We use it to override the
    #: default labels for the named columns
    verbose_names: Final[dict[str, str]] = {
        "title": "Project Name",
        "machine_name": "Machine Name",
        "latest_version": "Latest Version",
        "latest_version_date": "Import Date",
    }
    #: A dict of column names to alignment ("left", "right", "center")
    alignment: Final[dict[str, str]] = {
        "title": "left",
        "description": "left",
        "classifiers": "left",
        "machine_name": "left",
        "latest_version": "left",
        "latest_version_date": "left",
    }

    actions: Final[list[RowActionButton]] = [
        LatestVersionButton(),
        RowModelUrlButton(
            attribute="get_absolute_url", text="View", color="outline-secondary"
        ),
        RowEditButton(permission="sphinxhostingcore.change_project", color="azure"),
    ]

    def render_latest_version_column(self, row: Project, _: str) -> str:
        """
        Render our ``latest_version`` column.  This is the version string of the
        :py:class:`sphinx_hosting.models.Version` that has the most recent
        :py:attr:`sphinx_hosting.models.Version.modified` timestamp.

        If there are not yet any :py:class:`sphinx_hosting.models.Version` instances for
        this project, return empty string.

        Args:
            row: the ``Project`` we are rendering
            colunn: the name of the column to render

        Returns:
            The version string of the most recently published version, or empty
            string.

        """
        version = row.latest_version
        if version:
            return version.version  # type: ignore[attr-defined]
        return ""

    def render_latest_version_date_column(self, row: Project, _: str) -> str:
        """
        Render our ``latest_version_date`` column.  This is the last modified
        date of the :py:class:`sphinx_hosting.models.Version` that has the most
        recent :py:attr:`sphinx_hosting.models.Version.modified` timestamp.

        If there are not yet any :py:class:`sphinx_hosting.models.Version`
        instances for this project, return empty string.

        Args:
            row: the ``Project`` we are rendering
            colunn: the name of the column to render

        Returns:
            The of the most recently published version, or empty
            string.

        """
        version = row.latest_version
        if version:
            return self.render_datetime_type_column(version.modified)  # type: ignore[attr-defined]
        return ""

    def render_classifiers_column(self, row: Project, _: str) -> str:
        """
        Render our ``classifiers`` column.

        Args:
            row: the ``Project`` we are rendering
            colunn: the name of the column to render

        Returns:
            A ``<br>`` separated list of classifier names

        """
        return "<br>".join(row.classifiers.values_list("name", flat=True))  # type: ignore[attr-defined]

    def filter_classifiers_column(self, qs: QuerySet, _: str, value: str) -> QuerySet:
        """
        Filter our results by the ``value``, a comma separated list of
        :py:class:`sphinx_hosting.models.Classifier` names.

        Args:
            qs: the current :py:class:`QuerySet`
            colunn: the name of the column to filter on
            value: a comma-separated list of classifier names

        Returns:
            A :py:class:`QuerySet` filtered for rows that contain the selected
            classifiers.

        """
        classifier_ids = value.split(",")
        return qs.filter(classifiers__id__in=classifier_ids).distinct()


class ProjectVersionTable(BasicModelTable):
    """
    Displays a `dataTable <https://datatables.net>`_ of our
    :py:class:`sphinx_hosting.models.Version` instances for a particular
    :py:class:`sphinx_hosting.models.Project`.

    It's used as a the main widget in by :py:class:`ProjectVersionTableWidget`.
    """

    model: type[Model] = Version

    #: Show this many rows per page
    page_length: int = 25
    #: Set to ``True`` to stripe our table rows
    striped: bool = True
    actions: bool = True

    #: A list of fields that we will list as columns.  These are either fields
    #: on our :py:attr:`model`, or defined as ``render_FIELD_NAME_column`` methods
    #: on this class
    fields: Final[list[str]] = [
        "version",
        "num_pages",
        "num_images",
        "created",
        "modified",
    ]
    unsortable: Final[list[str]] = [
        "num_pages",
        "num_images",
    ]
    #: A list of names of columns that will will not be searched when doing a
    #: **global** search
    unsearchable: Final[list[str]] = [
        "num_pages",
        "num_images",
        "created",
        "modified",
    ]
    #: A dict of column name to column label.  We use it to override the
    #: default labels for the named columns
    verbose_names: Final[dict[str, str]] = {
        "title": "Version",
        "num_pages": "# Pages",
        "num_images": "# Images",
    }
    #: A dict of column names to alignment ("left", "right", "center")
    alignment: Final[dict[str, str]] = {
        "version": "left",
        "num_pages": "right",
        "num_images": "right",
        "created": "left",
        "modified": "left",
    }
    #: Sort so that the highest version number is on top
    sort_ascending: bool = False

    def __init__(self, *args, **kwargs) -> None:
        """
        One of our ``kwargs`` must be ``project_id``, the ``pk`` of the
        :py:class:`sphinx_hosting.models.Project` for which we want to list
        :py:class:`sphinx_hosting.models.Version` objects.

        This will get added to the :py:attr:`extra_data` dict in the ``kwargs``
        key, from which we reference it.
        """
        #: The pk of the :py:class:`sphinx_hosting.models.Project` for which to
        #: list versions
        self.project_id: int | None = kwargs.get("project_id")
        super().__init__(*args, **kwargs)
        if "project_id" in self.extra_data["kwargs"]:
            self.project_id = int(self.extra_data["kwargs"]["project_id"])

    def get_initial_queryset(self) -> QuerySet[Version]:
        """
        Filter our :py:class:`sphinx_hosting.models.Version` objects by
        :py:attr:`project_id`.

        Returns:
            A filtered :py:class:`QuerySet` on :py:class:`sphinx_hosting.models.Version`

        """
        return (
            super()
            .get_initial_queryset()
            .filter(project_id=self.project_id)
            .order_by("-version")
        )

    def render_num_pages_column(self, row: Version, _: str) -> str:
        """
        Render our ``num_pages`` column.  This is the number of
        :py:class:`sphinx_hosting.models.SphinxPage` objects imported for this
        version.

        Args:
            row: the ``Version`` we are rendering
            colunn: the name of the column to render

        Returns:
            The number of pages for this version.

        """
        return row.pages.count()

    def render_num_images_column(self, row: Version, _: str) -> str:
        """
        Render our ``num_images`` column.  This is the number of
        :py:class:`sphinx_hosting.models.SphinxImage` objects imported for this
        version.

        Args:
            row: the ``Version`` we are rendering
            colunn: the name of the column to render

        Returns:
            The number of images for this version.

        """
        return row.images.count()



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from django.urls import reverse
from wildewidgets import Widget, WidgetListLayout

from .forms import ProjectReadonlyUpdateForm
from .project_detail_layout import apply_project_detail_layout_builders
from .views import ProjectDetailView as BaseProjectDetailView
from .wildewidgets import (
    ProjectClassifierListWidget,
    ProjectDetailWidget,
    ProjectInfoWidget,
    ProjectRelatedLinksListWidget,
    ProjectVersionsTableWidget,
)

if TYPE_CHECKING:
    from django.contrib.auth.models import AbstractUser

    from .models import Project


class ProjectDetailView(BaseProjectDetailView):
    """
    Project detail view with host-project layout customization hooks.

    This subclass preserves the built-in ``ProjectDetailView`` layout while
    allowing host projects to mutate the live ``WidgetListLayout`` through
    ``PROJECT_DETAIL_LAYOUT_BUILDERS``.
    """

    def get_content(self) -> Widget:
        """
        Build the project detail layout and apply host-project extensions.

        Returns:
            The populated project detail ``WidgetListLayout``.

        Side Effects:
            May invoke configured project-detail layout builders that mutate the
            returned layout in place.

        """
        layout = WidgetListLayout(self.object.title)
        layout.add_widget(ProjectInfoWidget(self.object))
        layout.add_widget(
            ProjectDetailWidget(
                self.object, form=ProjectReadonlyUpdateForm(instance=self.object)
            )
        )
        layout.add_widget(
            ProjectRelatedLinksListWidget(queryset=self.object.related_links.all())
        )
        layout.add_widget(
            ProjectClassifierListWidget(queryset=self.object.classifiers.all())
        )
        layout.add_widget(ProjectVersionsTableWidget(project_id=self.object.pk))
        user = cast("AbstractUser", self.request.user)
        version = self.object.latest_version
        if version and version.head:
            layout.add_sidebar_link_button(
                "Read Docs",
                reverse(
                    "sphinx_hosting:sphinxpage--detail",
                    args=[
                        self.object.machine_name,
                        version.version,
                        version.head.relative_path,
                    ],
                ),
                color="orange",
                css_class="mb-3",
            )
        if user.has_perm("sphinxhostingcore.change_project"):
            layout.add_sidebar_link_button(
                "Edit Project",
                reverse(
                    "sphinx_hosting:project--update", args=[self.object.machine_name]
                ),
                color="azure",
            )
        apply_project_detail_layout_builders(
            request=self.request,
            user=user,
            project=cast("Project", self.object),
            layout=layout,
            view=self,
        )
        return layout



from django.urls import URLPattern, path, re_path

from .project_detail_view import ProjectDetailView
from .project_update_view import ProjectUpdateView
from .views import (
    ClassifierViewSet,
    GlobalSphinxPageSearchView,
    ProjectCreateView,
    ProjectDeleteView,
    ProjectListView,
    ProjectRelatedLinkCreateView,
    ProjectRelatedLinkDeleteView,
    ProjectRelatedLinkUpdateView,
    SphinxPageDetailView,
    VersionDeleteView,
    VersionDetailView,
    VersionMakeLatestView,
    VersionUploadView,
)
from .wildewidgets import ProjectClassifierSelectorWidget

app_name: str = "sphinx_hosting"

urlpatterns: list[URLPattern] = [
    path("", ProjectListView.as_view(), name="project--list"),
    path("project/", ProjectCreateView.as_view(), name="project--create"),
    path("project/<str:slug>/", ProjectDetailView.as_view(), name="project--detail"),
    path(
        "project/<str:slug>/update/",
        ProjectUpdateView.as_view(),
        name="project--update",
    ),
    path(
        "project/<str:slug>/upload/",
        VersionUploadView.as_view(),
        name="version--upload",
    ),
    path(
        "project/<str:slug>/delete/",
        ProjectDeleteView.as_view(),
        name="project--delete",
    ),
    path(
        "project/<str:project_slug>/link/",
        ProjectRelatedLinkCreateView.as_view(),
        name="projectrelatedlink--create",
    ),
    path(
        "project/link/<int:pk>/update/",
        ProjectRelatedLinkUpdateView.as_view(),
        name="projectrelatedlink--update",
    ),
    path(
        "project/link/<int:pk>/delete/",
        ProjectRelatedLinkDeleteView.as_view(),
        name="projectrelatedlink--delete",
    ),
    path(
        "project/<str:slug>/set-latest/",
        VersionMakeLatestView.as_view(),
        name="project--set-latest",
    ),
    path(
        "project/<str:project_slug>/<str:version>/",
        VersionDetailView.as_view(),
        name="version--detail",
    ),
    path(
        "project/<str:project_slug>/<str:version>/delete/",
        VersionDeleteView.as_view(),
        name="version--delete",
    ),
    re_path(
        r"project/(?P<project_slug>[-a-zA-Z0-9_.]+)/(?P<version>[^/]+)/(?P<path>.*)/",
        SphinxPageDetailView.as_view(),
        name="sphinxpage--detail",
    ),
    path("search/", GlobalSphinxPageSearchView.as_view(), name="search"),
]
urlpatterns += ClassifierViewSet(
    url_prefix="lookups", url_namespace=app_name
).get_urlpatterns()
urlpatterns += ProjectClassifierSelectorWidget.get_urlpatterns(url_namespace=app_name)



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, Any, Protocol, cast

from django.conf import settings as django_settings
from django.utils.module_loading import import_string

from .settings import PROJECT_DETAIL_LAYOUT_BUILDERS_SETTING

if TYPE_CHECKING:
    from django.contrib.auth.models import AbstractUser
    from django.http import HttpRequest
    from wildewidgets import WidgetListLayout

    from .models import Project

    class ProjectDetailLayoutView(Protocol):
        """
        Protocol for views that host project-detail layout builders.

        Both the read-only project detail view and the editable project update
        view satisfy this protocol.
        """

        request: HttpRequest

#: The setting path used in validation error messages for project-detail builders.
PROJECT_DETAIL_LAYOUT_BUILDERS_SOURCE: str = (
    "SPHINX_HOSTING_SETTINGS['PROJECT_DETAIL_LAYOUT_BUILDERS']"
)


class ProjectDetailLayoutBuilder(Protocol):
    """
    Protocol for a project-detail layout builder callable.

    The callable receives the live
    :py:class:`wildewidgets.WidgetListLayout` instance for a
    :py:class:`sphinx_hosting.views.ProjectDetailView` request and may mutate it
    in place.
    """

    def __call__(
        self,
        *,
        request: HttpRequest,
        user: AbstractUser,
        project: Project,
        layout: WidgetListLayout,
        view: ProjectDetailLayoutView,
    ) -> None:
        """
        Customize a project-detail page layout.

        Keyword Args:
            request: The current Django request.
            user: The authenticated user for the request.
            project: The project being displayed.
            layout: The live ``WidgetListLayout`` instance to mutate.
            view: The active project detail or update view instance.

        Side Effects:
            Mutates ``layout`` in place.

        """


@lru_cache(maxsize=64)
def _resolve_project_detail_layout_builders(
    paths: tuple[str, ...],
) -> tuple[ProjectDetailLayoutBuilder, ...]:
    """
    Resolve dotted import paths to project-detail layout builders.

    Args:
        paths: Dotted import paths for layout builder callables.

    Returns:
        A tuple of resolved project-detail layout builder callables.

    Raises:
        TypeError: One or more resolved objects is not callable.

    """
    builders: list[ProjectDetailLayoutBuilder] = []
    for path in paths:
        builder = import_string(path)
        if not callable(builder):
            msg = (
                f"{PROJECT_DETAIL_LAYOUT_BUILDERS_SOURCE} path '{path}' resolved to "
                f"'{type(builder).__name__}', which is not callable."
            )
            raise TypeError(msg)
        builders.append(cast("ProjectDetailLayoutBuilder", builder))
    return tuple(builders)


def get_project_detail_layout_builders() -> tuple[ProjectDetailLayoutBuilder, ...]:
    """
    Return configured project-detail layout builders.

    Returns:
        A tuple of resolved project-detail layout builder callables.

    Raises:
        TypeError: ``PROJECT_DETAIL_LAYOUT_BUILDERS`` is not a list of
            dotted-path strings.

    """
    app_settings: dict[str, Any] = getattr(
        django_settings, "SPHINX_HOSTING_SETTINGS", {}
    )
    raw_paths = app_settings.get(PROJECT_DETAIL_LAYOUT_BUILDERS_SETTING, [])
    if raw_paths is None:
        return ()
    if not isinstance(raw_paths, list):
        msg = (
            f"{PROJECT_DETAIL_LAYOUT_BUILDERS_SOURCE} must be a list of "
            "dotted-path strings."
        )
        raise TypeError(msg)
    for index, path in enumerate(raw_paths):
        if not isinstance(path, str):
            msg = (
                f"{PROJECT_DETAIL_LAYOUT_BUILDERS_SOURCE}[{index}] must be a "
                f"dotted-path string, got '{type(path).__name__}'."
            )
            raise TypeError(msg)
    return _resolve_project_detail_layout_builders(tuple(raw_paths))


def apply_project_detail_layout_builders(
    *,
    request: HttpRequest,
    user: AbstractUser,
    project: Project,
    layout: WidgetListLayout,
    view: ProjectDetailLayoutView,
) -> None:
    """
    Apply configured project-detail layout builders to ``layout``.

    Keyword Args:
        request: The current Django request.
        user: The authenticated user for the request.
        project: The project being displayed.
        layout: The live ``WidgetListLayout`` instance to mutate.
        view: The active project detail or update view instance.

    Side Effects:
        Allows configured builder callables to mutate ``layout`` in place.

    """
    for builder in get_project_detail_layout_builders():
        builder(
            request=request,
            user=user,
            project=project,
            layout=layout,
            view=view,
        )
